
import pytest

import pandas as pd
import numpy as np

from pyrle import Rle
from io import StringIO

from pyrle import PyRles
from pyrle.methods import coverage, _to_ranges, to_ranges

import pyranges as pr


@pytest.fixture
def chipseq_dataset():

    gr = pr.load_dataset("chipseq")

    return gr


@pytest.fixture
def teensy():

    c = """Chromosome Start End Name Score Strand
chr2 13611 13636 U0 2 -
chr2 32620 32645 U0 4 -
chr2 33241 33266 U0 1 +
chr2 788268 788293 U0 0 -
chr2 1150665 1150690 U0 -1 -"""

    return pr.PyRanges(pd.read_table(StringIO(c), sep="\s+"))


# def test_roundtrip_to_ranges_grles(chipseq_dataset):

#     cv = PyRles(chipseq_dataset)

#     chipseq_dataset.df.to_csv("input_dataset.txt", sep=" ")
#     gr = to_ranges(cv)

#     print(chipseq_dataset)
#     print(gr)
#     gr.df.to_csv("result.txt", sep=" ")

#     assert list(gr.df.Start) == list(chipseq_dataset.df.Start)
#     assert list(gr.df.Chromosome) == list(chipseq_dataset.df.Chromosome)
#     assert list(gr.df.End) == list(chipseq_dataset.df.End)


def test_roundtrip_to_ranges_single_rle_teensy(teensy):

    cv = coverage(teensy, value_col="Score")

    starts, ends, values = _to_ranges(cv)

    df = pd.concat([pd.Series(a) for a in [starts, ends, values]], axis=1)
    df.columns = "Start End Score".split()
    df.insert(0, "Chromosome", "chr2")

    gr = pr.PyRanges(df)

    # print(", ".join(str(i) for i in starts))
    # print(", ".join(str(f) for f in ends))
    assert list(starts) == [0, 13611, 13636, 32620, 32645, 33241, 33266, 1150665]
    assert list(ends) == [13611, 13636, 32620, 32645, 33241, 33266, 1150665, 1150690]


@pytest.fixture
def teensy_duplicated():

    "Contains duplicates"

    c = """Chromosome     Start       End Name  Score Strand
180       chr2  42058716  42058741   U0      0      +
181       chr2  42130511  42130536   U0      0      +
182       chr2  42593165  42593190   U0      0      -
183       chr2  42593165  42593190   U0      0      -
184       chr2  42635413  42635438   U0      0      -
185       chr2  43357333  43357358   U0      0      +
186       chr2  43854685  43854710   U0      0      +"""

    df = pd.read_table(StringIO(c), sep="\s+")

    return pr.PyRanges(df)



@pytest.fixture
def expected_result_teensy_duplicated():

    "Contains duplicates"

    c = """Chromosome     Start       End Name  Score Strand
180       chr2  42058716  42058741   U0      1      +
181       chr2  42130511  42130536   U0      1      +
183       chr2  42593165  42593190   U0      2      -
184       chr2  42635413  42635438   U0      1      -
185       chr2  43357333  43357358   U0      1      +
186       chr2  43854685  43854710   U0      1      +"""

    df = pd.read_table(StringIO(c), sep="\s+")

    return pr.PyRanges(df)


def test_roundtrip_to_ranges_single_rle_teensy_duplicated(teensy_duplicated, expected_result_teensy_duplicated):

    gr = teensy_duplicated
    cv = coverage(teensy_duplicated)
    # print(cv.values)

    starts, ends, values = _to_ranges(cv)

    # print(gr)
    # print(pr.PyRanges(gr.df.drop_duplicates()))
    # print("len(starts)", len(starts))
    # print("starts")
    # print(starts[:5])
    # print(starts[-5:])
    # print("ends")
    # print(ends[:5])
    # print(ends[-5:])

    assert list(starts) == [0, 42058716, 42058741, 42130511, 42130536, 42593165, 42593190, 42635413, 42635438, 43357333, 43357358, 43854685]
    assert list(ends) == [42058716, 42058741, 42130511, 42130536, 42593165, 42593190, 42635413, 42635438, 43357333, 43357358, 43854685, 43854710]


@pytest.fixture
def expected_result_single_chromosome(chipseq_dataset):

    return pr.PyRanges(chipseq_dataset["chr2"].df.drop_duplicates())


@pytest.fixture
def overlapping_gr():

    c = """Chromosome      Start        End Name  Score Strand
chr2  1  7   U0      0      -
chr2  4  10   U0      0      +"""

    df = pd.read_table(StringIO(c), sep="\s+")

    return pr.PyRanges(df)


def test_roundtrip_to_ranges_single_rle_overlapping(overlapping_gr):

    gr = overlapping_gr
    print(gr)
    cv = coverage(gr)
    print(cv)

    starts, ends, values = _to_ranges(cv)

    df = pd.concat([pd.Series(a) for a in [starts, ends, cv.values]], axis=1)
    df.columns = "Start End Score".split()
    df.insert(0, "Chromosome", "chr2")
    print(pr.PyRanges(df))
    # print(gr)
    # print(pr.PyRanges(gr.df.drop_duplicates()))
    # print("len(starts)", len(starts))
    # print("starts")
    # print(starts[:5])
    # print("ends")
    # print(ends[:5])


    assert list(starts) == [0, 1, 4, 7]
    assert list(ends) == [1, 4, 7, 10]

    # assert (starts == expected_result_problematic_gr.df.Start).all()
    # assert (ends == expected_result_problematic_gr.df.End).all()
    # assert (scores == expected_result_problematic_gr.df.Score).all()

# gr = PyRanges('toyChr',IRanges(cumsum(c(0,runLength(toyData)[-nrun(toyData)])),
#                               width=runLength(toyData)),
#              toyData = runValue(toyData))




# def test_roundtrip_to_ranges_single_chromosome(chipseq_dataset, expected_result_single_chromosome):

#     chr2 = chipseq_dataset["chr2"]
#     print(chr2)
#     print(expected_result_single_chromosome)

#     cv = coverage(chr2)
#     print(cv.values)

#     starts, ends, scores = _to_ranges(cv)

#     print(chr2.df.loc[list(range(600, 640))])
#     print(starts[183])
#     print(ends[183])

#     # gr = pr.PyRanges(pd.concat([chr2.df.head(5), chr2.df.tail(5)]))
#     # print(gr)
#     print("len(starts)", len(starts))
#     print("starts")
#     print(starts[:5])
#     print(starts[-5:])
#     print("ends")
#     print(ends[:5])
#     print(ends[-5:])
#     print("scores")
#     print(scores[:5])
#     print(scores[-5:])

#     assert list(starts) == list(expected_result_single_chromosome.df.Start)
#     assert (ends == expected_result_single_chromosome.df.End).all()
#     assert (scores == expected_result_single_chromosome.df.Score).all()



def test_create_grles(chipseq_dataset):

    grles = PyRles(chipseq_dataset)

    print(grles)



def test_create_stranded_grles(chipseq_dataset):


    grles = PyRles(chipseq_dataset, stranded=True)

    print(grles)


def test_create_stranded_grles_multicpu(chipseq_dataset):


    grles = PyRles(chipseq_dataset, stranded=True, n_jobs=5)

    print(grles)


# def test_grange_to_grle_and_back():







@pytest.fixture
def expected_result_coverage():

    runs = """9739215      25  463205      25 3069430      25    9143      25  993038 25  142071      25  260968      25   71512      25   18072      25 103292 25""".split()
    runs = [int(s) for s in runs]
    values = [0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1.]

    return runs, values


@pytest.fixture
def df():

    c = """chr21	9739215	9739240	U0	0	-
chr21	10202445	10202470	U0	0	+
chr21	13271900	13271925	U0	0	-
chr21	13281068	13281093	U0	0	-
chr21	14274131	14274156	U0	0	-
chr21	14416227	14416252	U0	0	+
chr21	14677220	14677245	U0	0	-
chr21	14748757	14748782	U0	0	+
chr21	14766854	14766879	U0	0	+
chr21	14870171	14870196	U0	0	-"""

    return pd.read_table(StringIO(c), header=None, names="Chromosome Start End Name Score Strand".split())


@pytest.fixture
def grle1():

    r1 = Rle([1, 5, 10], [1, 2, 3])
    r2 = Rle([4, 7, 9], [0.01, 0.02, 0.03])

    d = {"chr1": r1, "chr2": r2}
    print(d)
    g = PyRles(d)
    print(g)
    return g

@pytest.fixture
def grle1_stranded(grle1):

    return PyRles({("chr1", "+"): grle1["chr1"], ("chr1", "-"): grle1["chr1"], ("chr2", "-"): grle1["chr2"]})


@pytest.fixture()
def grle2():

    r1 = Rle([1, 2, 3], [1, 2, 3])
    r2 = Rle([5, 4, 2], [-0.1, -0.2, -0.3])

    d = {"chr1": r1, "chr2": r2}

    return PyRles(d)

@pytest.fixture
def expected_result():

    r1 = Rle([1, 2, 3, 10], [2, 4, 5, 3])
    r2 = Rle([4, 1, 4, 2, 9], [-0.09, -0.08, -0.18, -0.28, 0.03])

    d = {"chr1": r1, "chr2": r2}

    return PyRles(d)


@pytest.fixture
def d1():

    r1 = Rle([1, 5, 10], [1, 2, 3])
    r2 = Rle([4, 7, 9], [0.01, 0.02, 0.03])
    print(r1)

    d = {"chr1": r1, "chr2": r2}

    return d


def test_create_PyRles(d1):

    # No errors
    result = PyRles(d1)



def test_add_PyRles(grle1, grle2, expected_result):

    result = grle1.add(grle2)
    print(result)
    print("----")
    print(expected_result)

    assert result == expected_result


def test_create_stranded_grles(grle1_stranded):

    result = to_ranges(grle1_stranded)




# df = read.table(text="Chromosome Start End Name Score Strand
# 625 chr2 175474407 175474432 U0 0 -
# 626 chr2 175474427 175474452 U0 0 +", sep=" ")

# ir = IRanges(start=df$Start, end=df$End)

# cv = coverage(ir)


# ir2 = IRanges(cumsum(c(0,runLength(cv)[-nrun(cv)])),
#               width=runLength(cv))



import pytest

import pandas as pd
import numpy as np

from io import StringIO

from pyrle import PyRles

import pyranges as pr

@pytest.fixture
def chip():

    c = """Chromosome Start End Strand
chr1 5 7 +
chr1 3 10 -
chr2 1 3 +"""

    return PyRles(pd.read_table(StringIO(c), sep="\s+"))


@pytest.fixture
def background():

    c = """Chromosome Start End Strand
chr1 1 4 +
chr1 2 5 -
chr2 0 1 +"""

    return PyRles(pd.read_table(StringIO(c), sep="\s+"))


@pytest.fixture
def expected_result():

    c = """Chromosome  Start  End  Score
chr1      0    1    0.0
chr1      1    2   -1.0
chr1      2    3   -2.0
chr1      3    4   -1.0
chr1      4    5    0.0
chr1      5    7    2.0
chr1      7   10    1.0
chr2      0    1   -1.0
chr2      1    3    1.0"""

    grles =  PyRles(pd.read_table(StringIO(c), sep="\s+"), value_col="Score")

    return grles


def test_subtraction(chip, background, expected_result):

    result = chip - background

    print("result\n", result, "\n")
    print("expected_result\n", expected_result)
    assert result == expected_result




def test_subtraction_multicpu(chip, background, expected_result):

    result = chip.sub(background, n_jobs=5)

    print("result\n", result, "\n")
    print("expected_result\n", expected_result)
    assert result == expected_result



from pyrle.src.rle import sub_rles, add_rles, mul_rles, div_rles_zeroes, div_rles_nonzeroes
from pyrle.src.coverage import _remove_dupes

import pandas as pd
import numpy as np

from tabulate import tabulate

from numbers import Number

class Rle:

    def __init__(self, runs, values):
        assert len(runs) == len(values)

        runs = np.array(runs, dtype=np.int)
        values = np.array(values, dtype=np.double)
        s = pd.Series(values, dtype=np.double)

        if (s.shift() == s).any():
            runs, values = _remove_dupes(runs, values, len(values))

        self.runs = runs
        self.values = values


    def __radd__(self, other):

        return Rle(self.runs, self.values + other)

    def __rmul__(self, other):

        return Rle(self.runs, self.values * other)

    def __rsub__(self, other):

        return Rle(self.runs, other - self.values)

    def __rtruediv__(self, other):

        return Rle(self.runs, other / self.values)

    def __add__(self, other):

        if isinstance(other, Number):
            return Rle(self.runs, self.values + other)

        runs, values = add_rles(self.runs, self.values, other.runs, other.values)
        return Rle(runs, values)


    def __sub__(self, other):

        if isinstance(other, Number):
            return Rle(self.runs, self.values - other)

        runs, values = sub_rles(self.runs, self.values, other.runs, other.values)
        return Rle(runs, values)

    def __mul__(self, other):

        if isinstance(other, Number):
            return Rle(self.runs, self.values * other)

        runs, values = mul_rles(self.runs, self.values, other.runs, other.values)
        return Rle(runs, values)

    __rmul__ = __mul__

    def __truediv__(self, other):

        if isinstance(other, Number):
            return Rle(self.runs, self.values / other)

        if (other.values == 0).any():
            runs, values = div_rles_zeroes(self.runs, self.values, other.runs, other.values)
        else:
            runs, values = div_rles_nonzeroes(self.runs, self.values, other.runs, other.values)

        return Rle(runs, values)

    def __eq__(self, other):
        runs_equal = np.equal(self.runs, other.runs).all()
        values_equal = np.allclose(self.values, other.values)
        return runs_equal and values_equal

    def __str__(self):

        if len(self.runs) > 10:
            runs = [str(i) for i in self.runs[:5]] + \
                [" ... "] + [str(i) for i in self.runs[-5:]]
            values = ["{0:.3f}".format(i) for i in self.values[:5]] + \
                    [" ... "] + ["{0:.3f}".format(i) for i in self.values[-5:]]
        else:
            runs = [str(i) for i in self.runs]
            values = ["{0:.3f}".format(i) for i in self.values]

        df = pd.Series(values).to_frame().T

        df.columns = list(runs)
        df.index = ["Values"]
        df.index.name = "Runs"

        outstr = tabulate(df, tablefmt='psql', showindex=True, headers="keys")
        length = np.sum(self.runs)
        elements = len(self.runs)
        info = "\nRle of length {} containing {} elements".format(str(length), str(elements))

        return outstr + info


    def __repr__(self):

        return str(self)



import pytest

import pandas as pd
import numpy as np

from io import StringIO

from pyrle import Rle
from pyrle.methods import _to_ranges, coverage

from pyranges import PyRanges

@pytest.fixture
def simple_rle():

    r = pd.Series([1, 2, 3, 4], dtype=np.int16)
    r1 = Rle(r, r)

    return r1

@pytest.fixture
def simple_rle2():

    r = pd.Series([1, 2, 3, 4], dtype=np.int16)
    r2 = Rle(r * 2, r * 2)

    return r2


@pytest.fixture
def shorter_rle():

    runs = pd.Series([1, 2, 3])
    values = pd.Series([1, 0, 1])
    r = Rle(runs, values)

    return r

@pytest.fixture
def long_rle():

    r = pd.Series([6, 5, 4, 3, 2, 1], dtype=np.int16)
    r2 = Rle(r, r)

    return r2

# numeric-Rle of length 20 with 4 runs
#   Lengths: 2 4 6 8
#   Values : 2 4 6 8
# numeric-Rle of length 20 with 7 runs
#   Lengths:  1  1  1  3  4  2  8
#   Values : -1  0 -2 -1 -2 -6 -8

def test_subtract_simple(simple_rle, simple_rle2):

    result = simple_rle - simple_rle2

    print(result.runs)
    print(result.values)

    expected_runs = np.array([1, 1, 1, 3, 4, 2, 8])
    expected_values = np.array([-1,  0, -2, -1, -2, -6,  -8])

    assert all(np.equal(result.runs, expected_runs))
    assert all(np.equal(result.values, expected_values))



def test_subtract_simple2(simple_rle, simple_rle2):

# numeric-Rle of length 20 with 7 runs
#   Lengths: 1 1 1 3 4 2 8
#   Values : 1 0 2 1 2 6 8

    result2 = simple_rle2 - simple_rle

    print(result2.runs)
    print(result2.values)
    expected_runs2 = [1, 1, 1, 3, 4, 2, 8]
    expected_values2 = [1, 0, 2, 1, 2, 6, 8]
    assert all(np.equal(result2.runs, expected_runs2))
    assert all(np.equal(result2.values, expected_values2))


def test_subtract_advanced(shorter_rle, long_rle):

# numeric-Rle of length 21 with 7 runs
#   Lengths:  1  2  8  4  3  2  1
#   Values : -5 -6 -5 -4 -3 -2 -1
    expected_runs = [1, 2, 8, 4, 3, 2, 1]
    expected_values = [-5, -6, -5, -4, -3, -2, -1]
    result = shorter_rle - long_rle
    print(result.runs)
    print(result.values)

    assert all(np.equal(result.runs, expected_runs))
    assert all(np.equal(result.values, expected_values))

# numeric-Rle of length 21 with 7 runs
#   Lengths: 1 2 8 4 3 2 1
#   Values : 5 6 5 4 3 2 1
def test_subtract_advanced(shorter_rle, long_rle):

    expected_runs = [1, 2, 8, 4, 3, 2, 1]
    expected_values2 = [5, 6, 5, 4, 3, 2, 1]
    result2 = long_rle - shorter_rle

    print("result2.runs", result2.runs)
    print("result2.values", result2.values)

    assert all(np.equal(result2.runs, expected_runs))
    assert all(np.equal(result2.values, expected_values2))


@pytest.fixture
def chip():

    c = """Chromosome Start End Strand
chr2 1 3 +"""

    return coverage(PyRanges(pd.read_table(StringIO(c), sep="\s+")))



@pytest.fixture
def background():

    c = """Chromosome Start End Strand
chr2 0 1 +"""

    return coverage(PyRanges(pd.read_table(StringIO(c), sep="\s+")))


def test_subtract_result_same_start(chip, background):

    print(chip)
    print(background)

    result = chip - background

    print(result)

    assert result == Rle([1, 2], [-1, 1])




@pytest.fixture
def chip_chr1():

    c = """Chromosome Start End Strand
chr1 5 7 +
chr1 3 10 -"""

    return coverage(PyRanges(pd.read_table(StringIO(c), sep="\s+")))


@pytest.fixture
def background_chr1():

    c = """Chromosome Start End Strand
chr1 1 4 +
chr1 2 5 -"""

    return coverage(PyRanges(pd.read_table(StringIO(c), sep="\s+")))



def test_subtract_result_chr1(chip_chr1, background_chr1):

    result = chip_chr1 - background_chr1

    print(result)

    assert len(result.values) == 7



import pytest

import pandas as pd
import numpy as np

from pyrle import Rle
from io import StringIO

from pyrle.methods import coverage

import pyranges as pr

# expected result for whole chr21 according to Rle
#   [1] 9739215      25  463205      25 3069430      25    9143      25  993038
#  [10]      25  142071      25  260968      25   71512      25   18072      25
#  [19]  103292      25  211969      25  275065      25  337961      25   28264
#  [28]      25   85316      25 1327122      25  942292      25  515321      25
#  [37]   51031      25  505279      25  106831      25  525295      25  428835
#  [46]      25 1075916      25   59077      25  397488      25  146032      25
#  [55]  241044      25   40723      25    6325      25  790758      25  636326
#  [64]      25  265715      25  707648      25  442479      25  623307      25
#  [73]   39301      25   56624      25   37674      25    6412      25   75632
#  [82]      25    7020      25   91867      25  516098      25  455342      25
#  [91]  207316      25   89376      25  220415      25   63873      25  178563
# [100]      25  208833      25  231384      25  233483      25  210876      25
# [109]   13625      25  698897      25    9427      25  199410      25 1334739
# [118]      25  181942      25   17516      25  186634      25   18445      25
# [127]   56362      25    6113      25    8690      25  314362      25  427565
# [136]      25  194922      25   25119      25   96491      25  814824      25
# [145]  329690      25   77988      25  715491      25   23244      25  305720
# [154]      25   45296      25  332040      25  174008      25  163489      25
# [163]   51692      25  622758      25  582426      25  167975      25   99400
# [172]      25   14499      25  481575      25  224239      25  109893      25
# [181]  229084      25  481166      25   76961      25  104924      25  262629
# [190]      25  123925      25   72451      25  423954      25  622114      25
# [199]  870208      25  291275      25   58970      25  189900      25  972143
# [208]      25  532150      25  157577      25  360979      25  122030      25
# [217]  365189      25 1376353      25  251038      25  338889      25


# def test_chr2_to_coverage():

#     gr = pr.load_dataset("chipseq")
#     gr_chr2 = gr["chr2"]

#     # print(gr_chr2)
#     # gr_chr2.df.to_csv("gr_chr2", sep=" ", index=False)

#     coverage(gr_chr2)

#     assert 0

@pytest.fixture
def simple():
  c = """Chromosome Start End Score
chr2      0    1   -1.0
chr2      1    3    1.0"""

  return pd.read_table(StringIO(c), sep="\s+")


def test_coverage_simple(simple):

    result = coverage(simple, value_col="Score")

    print(result)

    assert result == Rle([1, 2], [-1, 1])




@pytest.fixture()
def expected_result_coverage():

    runs = """9739215      25  463205      25 3069430      25    9143      25  993038 25  142071      25  260968      25   71512      25   18072      25 103292 25""".split()
    runs = [int(s) for s in runs]
    values = [0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1.]

    return runs, values


@pytest.fixture()
def df():

    c = """chr21	9739215	9739240	U0	0	-
chr21	10202445	10202470	U0	0	+
chr21	14416227	14416252	U0	0	+
chr21	14677220	14677245	U0	0	-
chr21	14748757	14748782	U0	0	+
chr21	13271900	13271925	U0	0	-
chr21	13281068	13281093	U0	0	-
chr21	14274131	14274156	U0	0	-
chr21	14870171	14870196	U0	0	-
chr21	14766854	14766879	U0	0	+"""

    return pd.read_table(StringIO(c), header=None, names="Chromosome Start End Name Score Strand".split())


def test_coverage(df, expected_result_coverage):

    expected_runs, expected_values = expected_result_coverage

    # result = coverage(df)
    result = coverage(df)

    print(result.runs)
    print(result.values)


    print(len(result.runs))
    print(len(result.values))

    print(expected_runs)
    print(expected_values)

    assert list(result.runs) == expected_runs
    assert list(result.values) == expected_values



@pytest.fixture
def supersimple_bed():

    c = """Start End
2 3"""

    return pd.read_table(StringIO(c), sep="\s+", header=0)


# def test_coverage(supersimple_bed):

#     result = coverage(supersimple_bed)

#     assert list(result.runs) == [2, 1]
#     assert list(result.values) == [0, 1]


# @pytest.fixture
# def empty_bed():

#     c = """Start End
# 2 2
# 4 4"""

#     return pd.read_table(StringIO(c), sep="\s+", header=0)


# @pytest.fixture()
# def expected_result_supersimple_bed2():

#     runs = np.array([2, 1, 1, 1], dtype=np.int)
#     values = np.array([0, 1, 0, 1], dtype=np.float)

#     return Rle(runs, values)


# def test_supersimple_bed2(empty_bed, expected_result_supersimple_bed2):

#     result = coverage(empty_bed)
#     print(result.runs)
#     print(result.values)

#     assert list(result.runs) == [2, 2]


@pytest.fixture
def simple_bed():

    c = """Start End Value
3 6 2.4
4 7 0.9
5 6 3.33"""

    return pd.read_table(StringIO(c), sep="\s+", header=0)


@pytest.fixture()
def expected_result_simple_bed():

    runs = np.array([3, 1, 1, 1, 1], dtype=np.int)
    values = np.array([0, 1, 2, 3, 1], dtype=np.float)

    return Rle(runs, values)

@pytest.fixture()
def expected_result_simple_bed_values():

    runs = np.array([3, 1, 1, 1, 1], dtype=np.int)
    values = np.array([0., 2.4, 3.3, 6.63, 0.9], dtype=np.float)

    return Rle(runs, values)


def test_simple_bed(simple_bed, expected_result_simple_bed):

    result = coverage(simple_bed)
    print(result.runs, expected_result_simple_bed.runs)
    print(result.values, expected_result_simple_bed.values)
    assert list(result.runs) == list(expected_result_simple_bed.runs)
    assert list(result.values) == list(expected_result_simple_bed.values)


def test_simple_bed_with_scores(simple_bed, expected_result_simple_bed_values):

    result = coverage(simple_bed, value_col="Value")
    print(result.runs, expected_result_simple_bed_values.runs)
    print(result.values, expected_result_simple_bed_values.values)
    assert list(result.runs) == list(expected_result_simple_bed_values.runs)
    assert np.allclose(result.values, expected_result_simple_bed_values.values)



from pyrle import Rle

from numbers import Number

# from joblib import Parallel, delayed

from pyrle import methods as m

from natsort import natsorted

try:
    dummy = profile
except:
    profile = lambda x: x


class PyRles():

    # @profile
    def __init__(self, ranges, n_jobs=1, stranded=False, value_col=None):

        # Construct PyRles from dict of rles
        if isinstance(ranges, dict):

            self.rles = ranges
            self.__dict__["stranded"] = True if len(list(ranges.keys())[0]) == 2 else False

        # Construct PyRles from ranges
        elif not stranded:

            try:
                df = ranges.df
            except:
                df = ranges

            grpby = list(df.groupby("Chromosome"))

            # if n_jobs > 1:
            #     _rles = Parallel(n_jobs=n_jobs)(delayed(m.coverage)(cdf, value_col=value_col) for _, cdf in grpby)
            # else:
            _rles = []
            for _, cdf in grpby:
                cv = m.coverage(cdf, value_col=value_col)
                _rles.append(cv)

            self.rles = {c: r for c, r in zip([c for c, _ in grpby], _rles)}
            self.__dict__["stranded"] = False

        else:

            try:
                df = ranges.df
            except:
                df = ranges

            cs = df["Chromosome Strand".split()].drop_duplicates()
            cs = list(zip(cs.Chromosome.tolist(), cs.Strand.tolist()))

            grpby = df.groupby("Chromosome Strand".split())

            # if n_jobs > 1:
            # _rles = Parallel(n_jobs=n_jobs)(delayed(m.coverage)(csdf, value_col=value_col) for cs, csdf in grpby)
            # else:
            _rles = []
            for cs, csdf in grpby:
                _rles.append(m.coverage(csdf, value_col=value_col))

            self.rles = {cs: r for cs, r in zip([cs for cs, _ in grpby], _rles)}
            self.__dict__["stranded"] = True

    def add(self, other, n_jobs=1):

        return m.binary_operation("add", self, other, n_jobs=n_jobs)

    def __add__(self, other, n_jobs=1):

        if isinstance(other, Number):
            return PyRles({cs: v + other for cs, v in self.items()})

        return m.binary_operation("add", self, other)

    def __radd__(self, other):

        return PyRles({cs: other + v for cs, v in self.items()})

    def sub(self, other, n_jobs=1):

        return m.binary_operation("sub", self, other, n_jobs=n_jobs)

    def __sub__(self, other):

        if isinstance(other, Number):
            return PyRles({cs: v - other for cs, v in self.items()})

        return m.binary_operation("sub", self, other)

    def __rsub__(self, other):

        return PyRles({cs: other - v for cs, v in self.items()})

    def mul(self, other, n_jobs=1):

        return m.binary_operation("mul", self, other, n_jobs=n_jobs)

    def __rmul__(self, other):

        return PyRles({cs: other * v for cs, v in self.items()})

    def __mul__(self, other):

        if isinstance(other, Number):
            return PyRles({cs: v * other for cs, v in self.items()})

        return m.binary_operation("mul", self, other)

    __rmul__ = __mul__

    def div(self, other, n_jobs=1):

        return m.binary_operation("div", self, other, n_jobs=n_jobs)

    def __rtruediv__(self, other):

        return PyRles({cs: other / v for cs, v in self.items()})

    def __truediv__(self, other):

        if isinstance(other, Number):
            return PyRles({cs: v / other for cs, v in self.items()})

        return m.binary_operation("div", self, other)


    def to_ranges(self):

        return m.to_ranges(self)

    @property
    def stranded(self):

        return len(self.keys()[0]) == 2

    def keys(self):

        return natsorted(list(self.rles.keys()))

    def values(self):

        return natsorted(list(self.rles.values()))

    def items(self):

        return natsorted(list(self.rles.items()))

    def add_pseudocounts(self, pseudo=0.01):

        for k, rle in self.items():

            rle.values.loc[rle.values == 0] = pseudo

    def __getitem__(self, key):

        key_is_string = isinstance(key, str)

        if key_is_string and self.stranded and key not in ["+", "-"]:
            plus = self.rles.get((key, "+"), Rle([1], [0]))
            rev = self.rles.get((key, "-"), Rle([1], [0]))

            return PyRles({(key, "+"): plus, (key, "-"): rev})

        # only return particular strand, but from all chromos
        elif key_is_string and self.stranded and key in ["+", "-"]:
            to_return = dict()
            for (c, s), rle in self.items():
                if s == key:
                    to_return[c, s] = rle

            return PyRles(to_return)

        elif key_is_string:

            return self.rles[key]

        elif len(key) == 2:

            return self.rles[key]

        else:
            raise IndexError("Must use chromosome, strand or (chromosome, strand) to get items from PyRles.")


    def __str__(self):

        keys = natsorted(self.rles.keys())
        stranded = True if len(list(keys)[0]) == 2 else False

        if not stranded:
            if len(keys) > 2:
                str_list = [keys[0],
                    str(self.rles[keys[0]]),
                    "...",
                    keys[-1],
                    str(self.rles[keys[-1]]),
                    "Unstranded PyRles object with {} chromosomes.".format(len(self.rles.keys()))]
            elif len(keys) == 2:
                str_list = [keys[0],
                            "-" * len(keys[0]),
                            str(self.rles[keys[0]]),
                            "",
                            keys[-1],
                            "-" * len(keys[-1]),
                            str(self.rles[keys[-1]]),
                            "Unstranded PyRles object with {} chromosomes.".format(len(self.rles.keys()))]
            else:
                str_list = [keys[0],
                            str(self.rles[keys[0]]),
                            "Unstranded PyRles object with {} chromosome.".format(len(self.rles.keys()))]

        else:
            if len(keys) > 2:
                str_list = [" ".join(keys[0]),
                    str(self.rles[keys[0]]),
                    "...",
                    " ".join(keys[-1]),
                    str(self.rles[keys[-1]]),
                    "PyRles object with {} chromosomes/strand pairs.".format(len(self.rles.keys()))]
            elif len(keys) == 2:
                str_list = [" ".join(keys[0]),
                            "-" * len(keys[0]),
                            str(self.rles[keys[0]]),
                            "",
                            " ".join(keys[-1]),
                            "-" * len(keys[-1]),
                            str(self.rles[keys[-1]]),
                            "PyRles object with {} chromosomes/strand pairs.".format(len(self.rles.keys()))]
            else:
                str_list = [" ".join(keys[0]),
                            str(self.rles[keys[0]]),
                            "PyRles object with {} chromosome/strand pairs.".format(len(self.rles.keys()))]

        outstr = "\n".join(str_list)

        return outstr


    def __eq__(self, other):

        if not self.rles.keys() == other.rles.keys():
            return False

        for c in self.rles.keys():

            if self.rles[c] != other.rles[c]:
                return False

        return True

    def __repr__(self):

        return str(self)


if __name__ == "__main__":

    # Must turn on macros in setup.py for line tracing to work
    "kernprof -l pyrle/rledict.py && python -m line_profiler coverage.py.lprof"

    from time import time
    import datetime

    import pandas as pd

    test_file = "/mnt/scratch/endrebak/genomes/chip/UCSD.Aorta.Input.STL002.bed.gz"

    df = pd.read_table(test_file, sep="\t", usecols=[0, 1, 2, 5], header=None,
                       names="Chromosome Start End Strand".split(), nrows=None)


    print("Done reading")
    start = time()

    result = PyRles(df, n_jobs=25)

    end = time()
    total = end - start

    total_dt = datetime.datetime.fromtimestamp(total)

    minutes_seconds = total_dt.strftime('%M\t%S\n')

    print(result)
    print(minutes_seconds)



import pytest

import pandas as pd
import numpy as np

from pyrle import Rle

@pytest.fixture
def simple_rle():

    r = pd.Series([1, 2, 3, 4], dtype=np.int16)
    r1 = Rle(r, r)

    return r1

@pytest.fixture
def simple_rle2():

    r = pd.Series([1, 2, 3, 4], dtype=np.int16)
    r2 = Rle(r * 2, r * 2)

    return r2

# numeric-Rle of length 21 with 8 runs
#   Lengths: 1 2 3 5 4 3 2 1
#   Values : 7 6 7 5 4 3 2 1

@pytest.fixture
def shorter_rle():

    runs = pd.Series([1, 2, 3])
    values = pd.Series([1, 0, 1])
    r = Rle(runs, values)

    return r

@pytest.fixture
def long_rle():

    r = pd.Series([6, 5, 4, 3, 2, 1], dtype=np.int16)
    r2 = Rle(r, r)

    return r2


def test_add_simple(simple_rle, simple_rle2):

    result = simple_rle + simple_rle2

    expected_runs = np.array([1, 1, 1, 3, 4, 2, 8])
    expected_values = np.array([3, 4, 6, 7, 10, 6, 8])

    print(result.runs)
    print(result.values)

    assert all(np.equal(result.runs, expected_runs))
    assert all(np.equal(result.values, expected_values))

    result2 = simple_rle2 + simple_rle

    print(result2.runs)
    print(result2.values)

    print(result2)

    assert all(np.equal(result2.runs, expected_runs))
    assert all(np.equal(result2.values, expected_values))


def test_add_advanced(shorter_rle, long_rle):

    result = shorter_rle + long_rle
    print("result runs", result.runs)
    print("result values", result.values)

    result2 = long_rle + shorter_rle

    print("result runs", result2.runs)
    print("result values", result2.values)

    expected_runs = np.array([1, 2, 3, 5, 4, 3, 2, 1])
    expected_values = np.array([7.0, 6.0, 7.0, 5.0, 4.0, 3.0, 2.0, 1.0])

    assert all(np.equal(result.runs, expected_runs))
    assert all(np.equal(result.values, expected_values))

    assert all(np.equal(result2.runs, expected_runs))
    assert all(np.equal(result2.values, expected_values))



# cython: infer_types=True

import numpy as np

cimport cython




@cython.boundscheck(False)
@cython.wraparound(False)
cpdef add_rles(long [::1] runs1, double [::1] values1, long [::1] runs2, double [::1] values2):

    cdef int x1 = 0
    cdef int x2 = 0
    cdef int xn = 0
    cdef int nr = 0
    cdef double nv = 0
    cdef double diff = 0
    cdef int l1 = len(runs1)
    cdef int l2 = len(runs2)
    cdef long r1 = runs1[x1]
    cdef long r2 = runs2[x2]
    nrs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.long)
    nvs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr

    while(x1 < l1 and x2 < l2):

        diff = r1 - r2
        nv = values1[x1] + values2[x2]
        if diff < 0:
            nr = r1
            r2 = r2 - r1
            x1 += 1
            if x1 < l1:
                r1 = runs1[x1]
        elif diff > 0:
            nr = r2
            r1 = r1 - r2
            x2 += 1
            if x2 < l2:
                r2 = runs2[x2]
        else:
            nr = r2
            x1 += 1
            x2 += 1
            if x1 < l1:
                r1 = runs1[x1]
            if x2 < l2:
                r2 = runs2[x2]

        # if the new value is the same as the old, merge the runs
        if xn > 0 and nv == nvs[xn - 1]:
            nrs[xn - 1] += nr
        else:
            nrs[xn] = nr
            nvs[xn] = nv
            xn += 1

    # Here we unwind the rest of the values that were not added because one Rle was longer than the other.
    # (If other had largest sum of lengths.)
    if x1 == l1 and not x2 == l2:

        # Have some values left in one rl from the previous comparison
        # which must be added before we move on
        if diff < 0:
            nrs[xn] = r2
            nvs[xn] += values2[x2]
            xn += 1
            x2 += 1
        # if the new value is same as the old, merge
        if 0 + values2[x2] == nv:
            nrs[xn - 1] += runs2[x2]
            x2 += 1
        # now the unwinding; add all the values missing from one Rle
        for i in range(x2, l2):
            nrs[xn] = runs2[i]
            nvs[xn] += values2[i]
            xn += 1
    # (If self had largest sum of lengths)
    elif x2 == l2 and not x1 == l1:

        if diff > 0:
            nrs[xn] = r1
            nvs[xn] += values1[x1]
            xn += 1
            x1 += 1

        if values1[x1] == nvs[xn - 1]:
            nrs[xn - 1] += runs1[x1]
            x1 += 1

        for i in range(x1, l1):
            nrs[xn] = runs1[i]
            nvs[xn] = values1[i]
            xn += 1

    # Must use resize because initial guess for array was likely way too large
    nrs_arr.resize(xn, refcheck=False)
    nvs_arr.resize(xn, refcheck=False)

    return nrs_arr, nvs_arr


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef sub_rles(long [::1] runs1, double [::1] values1, long [::1] runs2, double [::1] values2):

    cdef int x1 = 0
    cdef int x2 = 0
    cdef int xn = 0
    cdef int nr = 0
    cdef double nv = 0
    cdef double diff = 0
    cdef int l1 = len(runs1)
    cdef int l2 = len(runs2)
    cdef long r1 = runs1[x1]
    cdef long r2 = runs2[x2]
    nrs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.long)
    nvs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr

    while(x1 < l1 and x2 < l2):

        diff = r1 - r2
        nv = values1[x1] - values2[x2]
        if diff < 0:
            nr = r1
            r2 = r2 - r1
            x1 += 1
            if x1 < l1:
                r1 = runs1[x1]
        elif diff > 0:
            nr = r2
            r1 = r1 - r2
            x2 += 1
            if x2 < l2:
                r2 = runs2[x2]
        else:
            nr = r2
            x1 += 1
            x2 += 1
            if x1 < l1:
                r1 = runs1[x1]
            if x2 < l2:
                r2 = runs2[x2]

        # if the new value is the same as the old, merge the runs
        if xn > 0 and nv == nvs[xn - 1]:
            nrs[xn - 1] += nr
        else:
            nrs[xn] = nr
            nvs[xn] = nv
            xn += 1

    # Here we unwind the rest of the values that were not added because one Rle was longer than the other.
    # (If other had largest sum of lengths.)
    if x1 == l1 and not x2 == l2:

        # Have some values left in one rl from the previous comparison
        # which must be added before we move on
        if diff < 0:
            nrs[xn] = r2
            nvs[xn] -= values2[x2]
            xn += 1
            x2 += 1
        # if the new value is same as the old, merge
        if 0 - values2[x2] == nv:
            nrs[xn - 1] += runs2[x2]
            x2 += 1
        # now the unwinding; add all the values missing from one Rle
        for i in range(x2, l2):
            nrs[xn] = runs2[i]
            nvs[xn] -= values2[i]
            xn += 1
    # (If self had largest sum of lengths)
    elif x2 == l2 and not x1 == l1:

        if diff > 0:
            nrs[xn] = r1
            nvs[xn] += values1[x1]
            xn += 1
            x1 += 1

        if values1[x1] == nvs[xn - 1]:
            nrs[xn - 1] += runs1[x1]
            x1 += 1

        for i in range(x1, l1):
            nrs[xn] = runs1[i]
            nvs[xn] = values1[i]
            xn += 1

    # Must use resize because initial guess for array was likely way too large
    nrs_arr.resize(xn, refcheck=False)
    nvs_arr.resize(xn, refcheck=False)

    return nrs_arr, nvs_arr




@cython.boundscheck(False)
@cython.wraparound(False)
cpdef div_rles_nonzeroes(long [::1] runs1, double [::1] values1, long [::1] runs2, double [::1] values2):

    cdef int x1 = 0
    cdef int x2 = 0
    cdef int xn = 0
    cdef int nr = 0
    cdef double nv = 0
    cdef double diff = 0
    cdef int l1 = len(runs1)
    cdef int l2 = len(runs2)
    cdef long r1 = runs1[x1]
    cdef long r2 = runs2[x2]
    nrs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.long)
    nvs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr

    while(x1 < l1 and x2 < l2):

        diff = r1 - r2
        nv = values1[x1] / values2[x2]
        if diff < 0:
            nr = r1
            r2 = r2 - r1
            x1 += 1
            if x1 < l1:
                r1 = runs1[x1]
        elif diff > 0:
            nr = r2
            r1 = r1 - r2
            x2 += 1
            if x2 < l2:
                r2 = runs2[x2]
        else:
            nr = r2
            x1 += 1
            x2 += 1
            if x1 < l1:
                r1 = runs1[x1]
            if x2 < l2:
                r2 = runs2[x2]

        # if the new value is the same as the old, merge the runs
        if xn > 0 and nv == nvs[xn - 1]:
            nrs[xn - 1] += nr
        else:
            nrs[xn] = nr
            nvs[xn] = nv
            xn += 1

    # Here we unwind the rest of the values that were not added because one Rle was longer than the other.
    # (If other had largest sum of lengths.)
    if x1 == l1 and not x2 == l2:

        # Have some values left in one rl from the previous comparison
        # which must be added before we move on
        if diff < 0:
            nrs[xn] = r2
            nvs[xn] = 0 if values2[x2] else np.nan
            xn += 1
            x2 += 1

        for i in range(x2, l2):
            nv = 0 if values2[i] else np.nan

            if nv == nvs[xn -1]:
                nrs[xn - 1] += runs2[i]
            else:
                nrs[xn] = runs2[i]
                nvs[xn] = nv
                xn += 1
    # (If self had largest sum of lengths)
    elif x2 == l2 and not x1 == l1:

        if diff > 0:
            nrs[xn] = r1
            nvs[xn] = np.inf * np.sign(values1[x1])
            xn += 1
            x1 += 1

        for i in range(x1, l1):
            nv = np.inf * np.sign(values1[i]) if values1[i] else np.nan

            if nv == nvs[xn -1]:
                nrs[xn - 1] += runs1[i]
            else:
                nrs[xn] = runs1[i]
                nvs[xn] = nv
                xn += 1

    # Must use resize because initial guess for array was likely way too large
    nrs_arr.resize(xn, refcheck=False)
    nvs_arr.resize(xn, refcheck=False)

    return nrs_arr, nvs_arr


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef div_rles_zeroes(long [::1] runs1, double [::1] values1, long [::1] runs2, double [::1] values2):

    cdef int x1 = 0
    cdef int x2 = 0
    cdef int xn = 0
    cdef int nr = 0
    cdef double nv = 0
    cdef double diff = 0
    cdef int l1 = len(runs1)
    cdef int l2 = len(runs2)
    cdef long r1 = runs1[x1]
    cdef long r2 = runs2[x2]
    nrs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.long)
    nvs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr

    while(x1 < l1 and x2 < l2):

        diff = r1 - r2
        
        if values2[x2] != 0:
             nv = values1[x1] / values2[x2]
        else:
             nv = np.inf * np.sign(values1[x1])

        if diff < 0:
            nr = r1
            r2 = r2 - r1
            x1 += 1
            if x1 < l1:
                r1 = runs1[x1]
        elif diff > 0:
            nr = r2
            r1 = r1 - r2
            x2 += 1
            if x2 < l2:
                r2 = runs2[x2]
        else:
            nr = r2
            x1 += 1
            x2 += 1
            if x1 < l1:
                r1 = runs1[x1]
            if x2 < l2:
                r2 = runs2[x2]

        # if the new value is the same as the old, merge the runs
        if xn > 0 and nv == nvs[xn - 1]:
            nrs[xn - 1] += nr
        else:
            nrs[xn] = nr
            nvs[xn] = nv
            xn += 1

    # Here we unwind the rest of the values that were not added because one Rle was longer than the other.
    # (If other had largest sum of lengths.)
    if x1 == l1 and not x2 == l2:

        # Have some values left in one rl from the previous comparison
        # which must be added before we move on
        if diff < 0:
            nrs[xn] = r2
            nvs[xn] = 0 if values2[x2] else np.nan
            xn += 1
            x2 += 1

        for i in range(x2, l2):
            nv = 0 if values2[i] else np.nan

            if nv == nvs[xn -1]:
                nrs[xn - 1] += runs2[i]
            else:
                nrs[xn] = runs2[i]
                nvs[xn] = nv
                xn += 1
    # (If self had largest sum of lengths)
    elif x2 == l2 and not x1 == l1:

        if diff > 0:
            nrs[xn] = r1
            nvs[xn] = np.inf * np.sign(values1[x1])
            xn += 1
            x1 += 1

        for i in range(x1, l1):
            nv = np.inf * np.sign(values1[i]) if values1[i] else np.nan

            if nv == nvs[xn -1]:
                nrs[xn - 1] += runs1[i]
            else:
                nrs[xn] = runs1[i]
                nvs[xn] = nv
                xn += 1

    # Must use resize because initial guess for array was likely way too large
    nrs_arr.resize(xn, refcheck=False)
    nvs_arr.resize(xn, refcheck=False)

    return nrs_arr, nvs_arr




@cython.boundscheck(False)
@cython.wraparound(False)
cpdef mul_rles(long [::1] runs1, double [::1] values1, long [::1] runs2, double [::1] values2):

    cdef int x1 = 0
    cdef int x2 = 0
    cdef int xn = 0
    cdef int nr = 0
    cdef double nv = 0
    cdef double diff = 0
    cdef int l1 = len(runs1)
    cdef int l2 = len(runs2)
    cdef long r1 = runs1[x1]
    cdef long r2 = runs2[x2]
    nrs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.long)
    nvs_arr = np.zeros(len(runs1) + len(runs2), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr

    while(x1 < l1 and x2 < l2):

        diff = r1 - r2
        nv = values1[x1] * values2[x2]
        if diff < 0:
            nr = r1
            r2 = r2 - r1
            x1 += 1
            if x1 < l1:
                r1 = runs1[x1]
        elif diff > 0:
            nr = r2
            r1 = r1 - r2
            x2 += 1
            if x2 < l2:
                r2 = runs2[x2]
        else:
            nr = r2
            x1 += 1
            x2 += 1
            if x1 < l1:
                r1 = runs1[x1]
            if x2 < l2:
                r2 = runs2[x2]

        # if the new value is the same as the old, merge the runs
        if xn > 0 and nv == nvs[xn - 1]:
            nrs[xn - 1] += nr
        else:
                nrs[xn] = nr
                nvs[xn] = nv
                xn += 1

    # Here we unwind the rest of the values that were not added because one Rle was longer than the other.
    # (If other had largest sum of lengths.)
    if x1 == l1 and not x2 == l2:

        # Have some values left in one rl from the previous comparison
        # which must be added before we move on
        if diff < 0:
            nrs[xn] = r2
            nvs[xn] = 0
            x2 += 1


        xn += 1
        for i in range(x2, l2):
            nrs[xn - 1] += runs2[i]

    # (If self had largest sum of lengths)
    elif x2 == l2 and not x1 == l1:

        if diff > 0:
            nrs[xn] = r1
            nvs[xn] = 0
            x1 += 1

        xn += 1
        for i in range(x1, l1):
            nrs[xn - 1] += runs1[i]

        # Must use resize because initial guess for array was likely way too large
    nrs_arr.resize(xn, refcheck=False)
    nvs_arr.resize(xn, refcheck=False)

    return nrs_arr, nvs_arr




import numpy as np
import pandas as pd

cimport cython


# try:
#     dummy = profile
# except:
#     profile = lambda x: x
def insort(a, b, kind='mergesort'):
    # took mergesort as it seemed a tiny bit faster for my sorted large array try.
    c = np.concatenate((a, b)) # we still need to do this unfortunatly.
    c.sort(kind=kind)
    flag = np.ones(len(c), dtype=bool)
    np.not_equal(c[1:], c[:-1], out=flag[1:])
    return c[flag]


# @profile
@cython.boundscheck(False)
@cython.wraparound(False)
def _coverage(long [::1] starts, long [::1] ends, double [::1] values, int length):

    d = {}

    cdef int i = 0
    cdef int start
    cdef int end
    cdef double value

    # all_pos = insort(starts, ends)

    # # print("values", np.array(values))
    while i < length:
        start = starts[i]
        value = values[i]
        # # print("start", start)
        # # print("value", value)
        if not start in d:
            # # print("If starts!")
            d[start] = value
            # # print(d[start])
        else:
            d[start] = d[start] + value
            # # print("Else starts!")
            # # print(d[start])
        end = ends[i]
        value = values[i]
        i = i + 1
        if not end in d:
            d[end] = -value
        else:
            d[end] = d[end] - value

    if 0 not in d:
        d[0] = 0

    sorted_items = sorted(d.items())
    # # print("sorted_items", sorted_items)
    runs = pd.Series([r[0] for r in sorted_items])
    # # print("runs\n", runs)
    value_series = pd.Series([v[1] for v in sorted_items])
    # # print("value_series\n", value_series)

    first_value = value_series[0]
    value_series = value_series.cumsum().shift()
    value_series[0] = first_value

    shifted = runs.shift()
    shifted[0] = 0
    runs = (runs - shifted)

    # print("runs", runs)
    # print("value_series", value_series)

    if len(value_series) > 1 and first_value == value_series[1]:
        runs[1] += runs[0]
        value_series = value_series[1:]
        runs = runs[1:]

    cdef long[::1] _runs
    cdef double[::1] _vals

    _runs = runs.astype(long).values
    _vals = value_series.values

    i = 0
    cdef int counter = 0
    cdef double old_val = _vals[i]
    cdef int old_run = _runs[i]
    cdef int run

    nrs_arr = np.zeros(len(runs), dtype=np.long)
    nvs_arr = np.zeros(len(runs), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr


    # print("value_series", value_series)
    for i in range(1, len(value_series)):

        run = _runs[i]
        value = _vals[i]
        # print("old_val", old_val)
        # print("old_run", old_run)
        # print("run", run)
        # print("value", value)

        if value == old_val:
            # print("in equal")
            old_run += run
        else:
            # print("new")
            nrs[counter] = old_run
            nvs[counter] = old_val
            old_run = run
            old_val = value
            counter += 1

    if len(value_series) == 1:
        # print("len value series one")
        return runs.values, value_series.values

    if value == old_val:
        # print("value == old val")
        nrs[counter] = old_run
        nvs[counter] = old_val
        counter += 1

    return nrs_arr[:counter], nvs_arr[:counter]





@cython.boundscheck(False)
@cython.wraparound(False)
def _remove_dupes(long [::1] runs, double [::1] values, int length):

    cdef long[::1] _runs
    cdef double[::1] _vals

    _runs = runs
    _vals = values

    i = 0
    cdef int counter = 0
    cdef double old_val = _vals[i]
    cdef int old_run = _runs[i]
    cdef int run

    nrs_arr = np.zeros(len(runs), dtype=np.long)
    nvs_arr = np.zeros(len(runs), dtype=np.double)

    cdef long[::1] nrs
    cdef double[::1] nvs

    nrs = nrs_arr
    nvs = nvs_arr


    # print("values", values)
    for i in range(1, len(values)):

        run = _runs[i]
        value = _vals[i]
        # print("old_val", old_val)
        # print("old_run", old_run)
        # print("run", run)
        # print("value", value)

        if value == old_val:
            # print("in equal")
            old_run += run
        else:
            # print("new")
            nrs[counter] = old_run
            nvs[counter] = old_val
            old_run = run
            old_val = value
            counter += 1

    if len(values) == 1:
        # print("len value series one")
        return runs.values, values.values

    if value == old_val:
        # print("value == old val")
        nrs[counter] = old_run
        nvs[counter] = old_val
        counter += 1

    return nrs_arr[:counter], nvs_arr[:counter]





import pytest

import pandas as pd
import numpy as np

from pyrle import Rle

@pytest.fixture
def weird_rle():

    return Rle([10, 20, 30, 40], [1, 2, 3, 4])



@pytest.fixture
def weird_rle2():

    return Rle([1], [1])


@pytest.fixture
def expected_result_weird():

    return Rle([1, 99], [1, 0])


def test_weird(weird_rle, weird_rle2, expected_result_weird):

    result = weird_rle2 * weird_rle
    result2 = weird_rle * weird_rle2

    print(result)

    assert result == expected_result_weird
    assert result2 == expected_result_weird


@pytest.fixture
def rle1():

    return Rle([3, 9, 8, 1, 2], [5, 4, 8, 1, 3])



@pytest.fixture
def rle2():

    return Rle([8, 8, 7], [4.2, 8.0, 7.3])

@pytest.fixture
def expected_result():

    runs = [int(i) for i in "3    5    4    4    4    1    2".split()]
    values = [float(f) for f in "21 16.8   32   64 58.4  7.3 21.9".split()]

    return Rle(runs, values)

def test_mult(rle1, rle2, expected_result):

    res = rle1 * rle2

    assert res == expected_result




import pytest

import pandas as pd
import numpy as np

from pyrle import Rle

@pytest.fixture
def weird_rle():

    return Rle([10, 20, 30, 40], [1, 2, 3, 4])



@pytest.fixture
def weird_rle2():

    return Rle([1], [1])


@pytest.fixture
def expected_result_weird():

    return Rle([1, 99], [1, 0])

def test_weird(weird_rle, weird_rle2, expected_result_weird):

    result = weird_rle2 / weird_rle

    print(result)

    assert result == expected_result_weird


@pytest.fixture
def expected_result_weird_opposite():

    return Rle([1, 99], [1, np.inf])


def test_weird_opposite(weird_rle, weird_rle2, expected_result_weird_opposite):

    result = weird_rle / weird_rle2

    print(result)

    assert result == expected_result_weird_opposite


@pytest.fixture
def simple_rle():

    r = pd.Series([1, 2, 3, 4], dtype=np.int16)
    r1 = Rle(r, r)

    return r1

@pytest.fixture
def simple_rle2():

    r = pd.Series([1, 2, 3, 4], dtype=np.int16)
    r2 = Rle(r * 2, r * 2)

    return r2

# numeric-Rle of length 21 with 8 runs
#   Lengths: 1 2 3 5 4 3 2 1
#   Values : 7 6 7 5 4 3 2 1

@pytest.fixture
def shorter_rle():

    runs = pd.Series([1, 2, 3])
    values = pd.Series([1, 0, 1])
    r = Rle(runs, values)

    return r

@pytest.fixture
def long_rle():

    r = pd.Series([6, 5, 4, 3, 2, 1], dtype=np.int16)
    r2 = Rle(r, r)

    return r2

@pytest.fixture
def long_rle_neg_vals():

    r = pd.Series([6, 5, 4, 3, 2, 1], dtype=np.int16)
    r2 = Rle(r, r * -1)

    return r2

@pytest.fixture
def expected_result_div_simple():

    return Rle([1, 1, 1, 3, 4, 10], [0.5, 1, 0.5, 0.75, 0.66666667, 0.0])


def test_div_simple(simple_rle, simple_rle2, expected_result_div_simple):

    result = simple_rle / simple_rle2

    print(result)
    print(expected_result_div_simple)


    print(result.values)
    print(expected_result_div_simple.values)

    assert result == expected_result_div_simple



@pytest.fixture
def expected_result_div_simple_opposite():

    return Rle([1, 1, 1, 3, 4, 10], [2.0, 1.0, 2.0, 1.333333, 1.5, np.inf])


def test_div_simple_opposite(simple_rle, simple_rle2, expected_result_div_simple_opposite):

    result = simple_rle2 / simple_rle

    print(result.values)
    print(result.values)

    print(result.runs)

    assert result == expected_result_div_simple_opposite


@pytest.fixture
def expected_result_div_advanced():

    return Rle([1, 2, 3, 15], [0.16666667, 0., 0.16666667, 0.])


def test_div_advanced(shorter_rle, long_rle, expected_result_div_advanced):

    result = shorter_rle / long_rle
    print("result runs", result.runs)
    print("result values", result.values)

    assert result == expected_result_div_advanced




@pytest.fixture
def expected_result_div_advanced_opposite():

    return Rle([1, 2, 3, 15], [6, np.inf, 6, np.inf])

def test_div_advanced_opposite(shorter_rle, long_rle, expected_result_div_advanced_opposite):

    result = long_rle / shorter_rle
    print("result runs", result.runs)
    print("result values", result.values)

    assert result == expected_result_div_advanced_opposite



@pytest.fixture
def expected_result_div_advanced_opposite():

    return Rle([1, 2, 3, 15], [6, np.inf, 6, np.inf])



@pytest.fixture
def expected_result_div_advanced_neg_vals():

    return Rle([1, 2, 3, 15], [-0.16666667, 0., -0.16666667, 0.])


def test_div_advanced_neg_vals(shorter_rle, long_rle_neg_vals, expected_result_div_advanced_neg_vals):

    result = shorter_rle / long_rle_neg_vals
    print("result runs", result.runs)
    print("result values", result.values)

    assert result == expected_result_div_advanced_neg_vals


@pytest.fixture
def expected_result_div_advanced_opposite_neg_vals():

    return Rle([1, 2, 3, 15], [-6, -np.inf, -6, -np.inf])


def test_div_advanced_opposite_neg_vals(shorter_rle, long_rle_neg_vals, expected_result_div_advanced_opposite_neg_vals):

    result = long_rle_neg_vals / shorter_rle
    print("result runs", result.runs)
    print("result values", result.values)

    assert result == expected_result_div_advanced_opposite_neg_vals



import pandas as pd
import numpy as np

from pyrle import Rle
from pyrle.rledict import PyRles
from pyrle.src.coverage import _coverage

from natsort import natsorted

from sys import stderr


from collections import defaultdict

try:
    dummy = profile
except:
    profile = lambda x: x

def _merge_rles(rle):

    new_dict = {}
    dd = defaultdict(list)
    for chromosome, strand in rle.rles.keys():
        dd[chromosome].append(strand)

    for c, s in dd.items():
        if len(s) == 1:
            new_dict[c] = rle.rles[c, s[0]]
        else:
            new_dict[c] = rle.rles[c, "+"] + rle.rles[c, "-"]
            # print("c", c)
            # print("rle.rles[c, '+']\n", rle.rles[c, "+"])
            # print("rle.rles[c, '-']\n", rle.rles[c, "-"])

    return new_dict


def ensure_both_or_none_stranded(self, other):

    # means other not stranded
    if self.stranded:
        self.rles = _merge_rles(self)
    else:
        other.rles = _merge_rles(other)

    return self, other


def chromosomes_in_both_self_other(self, other):

    chromosomes_in_both = set(self.rles.keys()).intersection(other.rles.keys())
    chromosomes_in_self_not_other = set(self.rles.keys()) - set(other.rles.keys())
    chromosomes_in_other_not_self = set(other.rles.keys()) - set(self.rles.keys())

    # if chromosomes_in_self_not_other:
    #     print(", ".join(natsorted(chromosomes_in_self_not_other)) + " missing from other.", file=stderr)

    # if chromosomes_in_other_not_self:
    #     print(", ".join(natsorted(chromosomes_in_other_not_self)) + " missing from self.", file=stderr)

    return chromosomes_in_both, chromosomes_in_self_not_other, chromosomes_in_other_not_self


def binary_operation(operation, self, other, n_jobs=1):

    func = {"div": __div, "mul": __mul, "add": __add, "sub": __sub}[operation]

    if self.stranded != other.stranded:
        self, other = ensure_both_or_none_stranded(self, other)

    chromosomes_in_both, chromosomes_in_self_not_other, chromosomes_in_other_not_self = chromosomes_in_both_self_other(self, other)

    _rles = []
    for c in chromosomes_in_both:
        _rles.append(func(self.rles[c], other.rles[c]))


    rles = {c: r for c, r in zip(chromosomes_in_both, _rles)}

    for c in chromosomes_in_self_not_other:
        rles[c] = self.rles[c]

    for c in chromosomes_in_other_not_self:
        rles[c] = other.rles[c]

    return PyRles(rles)

def __add(self, other):

    return self + other

def __sub(self, other):

    return self - other

def __div(self, other):

    return self / other

def __mul(self, other):

    return self * other



def coverage(ranges, value_col=None):

    try:
        df = ranges.df
    except:
        df = ranges

    if value_col:
        values = df[value_col].astype(np.float64).values
    else:
        values = np.ones(len(df))

    # ndf = df["Start End Score".split()].sort_values("Start End".split())
    # else ValueError: buffer source array is read-only
    new_starts = df.Start.copy().values
    new_ends = df.End.copy().values

    runs, values = _coverage(new_starts, new_ends, values,
                             len(df))

    return Rle(runs, values)




def to_ranges(grles):

    from pyranges import PyRanges

    dfs = []
    if grles.stranded:

        for (chromosome, strand), rle in grles.items():
            starts, ends, values = _to_ranges(rle)
            df = pd.concat([pd.Series(r) for r in [starts, ends, values]], axis=1)
            df.columns = "Start End Score".split()
            df.insert(0, "Chromosome", chromosome)
            df.insert(df.shape[1], "Strand", strand)
            dfs.append(df)
    else:

        for chromosome, rle in grles.items():
            starts, ends, values = _to_ranges(rle)
            df = pd.concat([pd.Series(r) for r in [starts, ends, values]], axis=1)
            df.columns = "Start End Score".split()
            df.insert(0, "Chromosome", chromosome)
            dfs.append(df)

    return PyRanges(pd.concat(dfs))


def _to_ranges(rle):

    runs = pd.Series(rle.runs)
    starts = pd.Series([0] + list(runs)).cumsum()

    ends = starts + runs

    values = pd.Series(rle.values)

    start_idx = values[values.shift(-1) != values].index
    end_idx = values[values.shift(1) != values].index

    starts = starts.loc[start_idx]
    ends = ends.loc[end_idx]
    values = values[start_idx].reset_index(drop=True)

    return starts.astype(int).reset_index(drop=True), ends.astype(int).reset_index(drop=True), values
